#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>

#define LISTEN_PORT 3000
#define BUFSIZE 5000

int gate_sock;
int local_sock;

void usage(char *prog)
{
	printf("  Wingate redir - usage:\n");
	printf("    %s [IP WinGate] [host:port]\n",prog);
}

int connect_wg(char *ip)
{
	struct sockaddr_in sa;
	char buf[20];
	int nrec;

	if ((gate_sock=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0)
	{
		perror("socket");
		return 0;
	}

	sa.sin_family=AF_INET;
	sa.sin_port=htons(23);
	sa.sin_addr.s_addr=inet_addr(ip);
	memset(&sa.sin_zero,0,sizeof(sa.sin_zero));

	if (connect(gate_sock,(struct sockaddr *)&sa,sizeof(sa)) < 0)
	{
		perror("connect");
		return 0;
	}

	fprintf(stderr,"Connected. Waiting for wingate prompt...");
	do
	{
		memset(buf,0,sizeof(buf));
		nrec = recv(gate_sock,buf,sizeof(buf),0);
	}
	while (strstr(buf,"WinGate") == NULL);
	fprintf(stderr," ok!\n");

	return 1;
}

int open_local()
{
	int listener;
	struct sockaddr_in sa;
	int sin_size;

	if ((listener=socket(PF_INET,SOCK_STREAM,IPPROTO_TCP)) < 0)
	{
		perror("socket");
		return 0;
	}

	sa.sin_family=AF_INET;
	sa.sin_addr.s_addr=INADDR_ANY;
	sa.sin_port=htons(LISTEN_PORT);
	memset(&sa.sin_zero,0,sizeof(sa.sin_zero));

	if (bind(listener,(struct sockaddr *)&sa,sizeof(sa)) < 0)
	{
		perror("bind");
		return 0;
	}

	if (listen(listener,1)<0)
	{
		perror("listen");
		return 0;
	}

	sin_size = sizeof(sa);
	if ((local_sock=accept(listener,(struct sockaddr *)&sa,&sin_size))<0)
	{
		perror("accept");
		return 0;
	}

	close(listener);

	return 1;
}

int redirect(char *addr)
{
	fd_set fset;
	int max_fd;
	char buf[BUFSIZE];

	send(gate_sock,addr,strlen(addr),0);
	send(gate_sock,"\n",1,0);

	fprintf(stderr,"Waiting for connection to be initiated...");
	do
	{
		int nrec;
		memset(buf,0,sizeof(buf));
		nrec = recv(gate_sock,buf,sizeof(buf),0);
	}
	while (strstr(buf,"Connected") == NULL);
	fprintf(stderr," ok!\n");

	max_fd=(local_sock>gate_sock)?local_sock:gate_sock;
	max_fd++;

	while (1)
	{
		FD_ZERO(&fset);
		FD_SET(local_sock,&fset);
		FD_SET(gate_sock,&fset);

		if (select(max_fd,&fset,NULL,NULL,NULL)<0)
		{
			perror("select");
			return 0;
		}
		if (FD_ISSET(local_sock,&fset))
		{
			int nrec;
			nrec=recv(local_sock,buf,sizeof(buf),0);
			if (nrec>0)
			{
				send(gate_sock,buf,nrec,0);
			}
		}
		if (FD_ISSET(gate_sock,&fset))
		{
			int nrec;
			nrec=recv(gate_sock,buf,sizeof(buf),0);
			if (nrec>0)
			{
				send(local_sock,buf,nrec,0);
			}
		}
	}

	return 1;
}

int main(int argc, char *argv[])
{
	if (argc<3)
	{
		usage(argv[0]);
		return 0;
	}
	if (!connect_wg(argv[1]))
	{
		return -1;
	}
	if (!open_local())
	{
		return -1;
	}
	if (!redirect(argv[2]))
	{
		return -1;
	}
	return 0;
}

